from __future__ import annotations

import stat

from dissect.cstruct import cstruct

ext_def = """
#define EXT2_SBOFF              1024        // offset to superblock
#define EXT2_FS_MAGIC           0xef53

#define EXT2_BAD_INO            1           // Bad blocks inode
#define EXT2_ROOT_INO           2           // Root inode
#define EXT4_USR_QUOTA_INO      3           // User quota inode
#define EXT4_GRP_QUOTA_INO      4           // Group quota inode
#define EXT2_BOOT_LOADER_INO    5           // Boot loader inode
#define EXT2_UNDEL_DIR_INO      6           // Undelete directory inode
#define EXT2_RESIZE_INO         7           // Reserved group descriptors inode
#define EXT2_JOURNAL_INO        8           // Journal inode

#define EXT2_NDIR_BLOCKS        12          // direct blocks in inode
#define EXT2_NIND_BLOCKS        3           // indirect blocks in inode

#define EXT2_NAME_LEN           255
#define EXT2_MIN_BLOCK_SIZE     1024
#define EXT2_MAX_BLOCK_SIZE     4096
#define EXT4_MAX_BLOCK_SIZE     65536

#define EXT2_FEATURE_COMPAT_DIR_PREALLOC        0x0001
#define EXT2_FEATURE_COMPAT_IMAGIC_INODES       0x0002
#define EXT3_FEATURE_COMPAT_HAS_JOURNAL         0x0004
#define EXT2_FEATURE_COMPAT_EXT_ATTR            0x0008
#define EXT2_FEATURE_COMPAT_RESIZE_INODE        0x0010
#define EXT2_FEATURE_COMPAT_DIR_INDEX           0x0020
#define EXT4_FEATURE_COMPAT_SPARSE_SUPER2       0x0200

#define EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER     0x0001
#define EXT2_FEATURE_RO_COMPAT_LARGE_FILE       0x0002
#define EXT2_FEATURE_RO_COMPAT_BTREE_DIR        0x0004
#define EXT2_FEATURE_RO_COMPAT_HUGE_FILE        0x0008
#define EXT4_FEATURE_RO_COMPAT_GDT_CSUM         0x0010
#define EXT4_FEATURE_RO_COMPAT_DIR_NLINK        0x0020
#define EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE      0x0040
#define EXT4_FEATURE_RO_COMPAT_QUOTA            0x0100
#define EXT4_FEATURE_RO_COMPAT_BIGALLOC         0x0200
#define EXT4_FEATURE_RO_COMPAT_METADATA_CSUM    0x0400
#define EXT4_FEATURE_RO_COMPAT_READONLY         0x1000
#define EXT4_FEATURE_RO_COMPAT_PROJECT          0x2000

#define EXT2_FEATURE_INCOMPAT_COMPRESSION       0x0001
#define EXT2_FEATURE_INCOMPAT_FILETYPE          0x0002
#define EXT3_FEATURE_INCOMPAT_RECOVER           0x0004
#define EXT3_FEATURE_INCOMPAT_JOURNAL_DEV       0x0008
#define EXT2_FEATURE_INCOMPAT_META_BG           0x0010
#define EXT4_FEATURE_INCOMPAT_EXTENTS           0x0040
#define EXT4_FEATURE_INCOMPAT_64BIT             0x0080
#define EXT4_FEATURE_INCOMPAT_MMP               0x0100
#define EXT4_FEATURE_INCOMPAT_FLEX_BG           0x0200
#define EXT4_FEATURE_INCOMPAT_EA_INODE          0x0400
#define EXT4_FEATURE_INCOMPAT_DIRDATA           0x1000
#define EXT4_FEATURE_INCOMPAT_CSUM_SEED         0x2000
#define EXT4_FEATURE_INCOMPAT_LARGEDIR          0x4000
#define EXT4_FEATURE_INCOMPAT_INLINE_DATA       0x8000          // data in inode
#define EXT4_FEATURE_INCOMPAT_ENCRYPT           0x10000         // >2GB or 3-lvl htree

#define EXT2_FT_FMT                             0x0000F000
#define EXT2_FT_SOCK                            0x0000C000
#define EXT2_FT_LNK                             0x0000A000
#define EXT2_FT_REG                             0x00008000
#define EXT2_FT_BLK                             0x00006000
#define EXT2_FT_DIR                             0x00004000
#define EXT2_FT_CHR                             0x00002000
#define EXT2_FT_FIFO                            0x00001000

#define EXT2_SECRM_FL                           0x00000001      // Secure deletion
#define EXT2_UNRM_FL                            0x00000002      // Undelete
#define EXT2_COMPR_FL                           0x00000004      // Compress file
#define EXT2_SYNC_FL                            0x00000008      // Synchronous updates
#define EXT2_IMMUTABLE_FL                       0x00000010      // Immutable file
#define EXT2_APPEND_FL                          0x00000020      // writes to file may only append
#define EXT2_NODUMP_FL                          0x00000040      // do not dump file
#define EXT2_NOATIME_FL                         0x00000080      // do not update atime
#define EXT2_DIRTY_FL                           0x00000100
#define EXT2_COMPRBLK_FL                        0x00000200      // One or more compressed clusters
#define EXT2_NOCOMP_FL                          0x00000400      // Don't compress
#define EXT2_ECOMPR_FL                          0x00000800      // Compression error
#define EXT4_ENCRYPT_FL                         0x00000800      // encrypted file
#define EXT2_BTREE_FL                           0x00001000      // btree format dir
#define EXT4_INDEX_FL                           0x00001000      // hash-indexed directory
#define EXT2_IMAGIC_FL                          0x00002000      // AFS directory
#define EXT2_JOURNAL_DATA_FL                    0x00004000      // file data should be journaled
#define EXT2_NOTAIL_FL                          0x00008000      // file tail should not be merged
#define EXT2_DIRSYNC_FL                         0x00010000      // dirsync behaviour (directories only)
#define EXT2_TOPDIR_FL                          0x00020000      // Top of directory hierarchies
#define EXT4_HUGE_FILE_FL                       0x00040000      // Set to each huge file
#define EXT4_EXTENTS_FL                         0x00080000      // Inode uses extents
#define EXT4_EA_INODE_FL                        0x00200000      // Inode used for large EA
#define EXT4_EOFBLOCKS_FL                       0x00400000      // Blocks allocated beyond EOF
#define EXT4_INLINE_DATA_FL                     0x10000000      // Inode has inline data.
#define EXT4_PROJINHERIT_FL                     0x20000000      // Create with parents projid
#define EXT2_RESERVED_FL                        0x80000000      // reserved

#define EXT2_FL_USER_VISIBLE                    0x304BDFFF      // User visible flags
#define EXT2_FL_USER_MODIFIABLE                 0x204BC0FF      // User modifiable flags

struct ext4_super_block {
    uint32      s_inodes_count;             /* Inodes count */
    uint32      s_blocks_count_lo;          /* Blocks count */
    uint32      s_r_blocks_count_lo;        /* Reserved blocks count */
    uint32      s_free_blocks_count_lo;     /* Free blocks count */
    uint32      s_free_inodes_count;        /* Free inodes count */
    uint32      s_first_data_block;         /* First Data Block */
    uint32      s_log_block_size;           /* Block size */
    uint32      s_log_cluster_size;         /* Allocation cluster size */
    uint32      s_blocks_per_group;         /* # Blocks per group */
    uint32      s_clusters_per_group;       /* # Clusters per group */
    uint32      s_inodes_per_group;         /* # Inodes per group */
    uint32      s_mtime;                    /* Mount time */
    uint32      s_wtime;                    /* Write time */
    uint16      s_mnt_count;                /* Mount count */
    uint16      s_max_mnt_count;            /* Maximal mount count */
    uint16      s_magic;                    /* Magic signature */
    uint16      s_state;                    /* File system state */
    uint16      s_errors;                   /* Behaviour when detecting errors */
    uint16      s_minor_rev_level;          /* minor revision level */
    uint32      s_lastcheck;                /* time of last check */
    uint32      s_checkinterval;            /* max. time between checks */
    uint32      s_creator_os;               /* OS */
    uint32      s_rev_level;                /* Revision level */
    uint16      s_def_resuid;               /* Default uid for reserved blocks */
    uint16      s_def_resgid;               /* Default gid for reserved blocks */
    uint32      s_first_ino;                /* First non-reserved inode */
    uint16      s_inode_size;               /* size of inode structure */
    uint16      s_block_group_nr;           /* block group # of this superblock */
    uint32      s_feature_compat;           /* compatible feature set */
    uint32      s_feature_incompat;         /* incompatible feature set */
    uint32      s_feature_ro_compat;        /* readonly-compatible feature set */
    char        s_uuid[16];                 /* 128-bit uuid for volume */
    char        s_volume_name[16];          /* volume name */
    char        s_last_mounted[64];         /* directory where last mounted */
    uint32      s_algorithm_usage_bitmap;   /* For compression */
    uint8       s_prealloc_blocks;          /* Nr of blocks to try to preallocate*/
    uint8       s_prealloc_dir_blocks;      /* Nr to preallocate for dirs */
    uint16      s_reserved_gdt_blocks;      /* Per group desc for online growth */
    // Journaling support valid if EXT4_FEATURE_COMPAT_HAS_JOURNAL set.
    char        s_journal_uuid[16];         /* uuid of journal superblock */
    uint32      s_journal_inum;             /* inode number of journal file */
    uint32      s_journal_dev;              /* device number of journal file */
    uint32      s_last_orphan;              /* start of list of inodes to delete */
    char        s_hash_seed[16];            /* HTREE hash seed */
    uint8       s_def_hash_version;         /* Default hash version to use */
    uint8       s_jnl_backup_type;
    uint16      s_desc_size;                /* size of group descriptor */
    uint32      s_default_mount_opts;
    uint32      s_first_meta_bg;            /* First metablock block group */
    uint32      s_mkfs_time;                /* When the filesystem was created */
    uint32      s_jnl_blocks[17];           /* Backup of the journal inode */
    // 64bit support valid if EXT4_FEATURE_COMPAT_64BIT
    uint32      s_blocks_count_hi;          /* Blocks count */
    uint32      s_r_blocks_count_hi;        /* Reserved blocks count */
    uint32      s_free_blocks_count_hi;     /* Free blocks count */
    uint16      s_min_extra_isize;          /* All inodes have at least # bytes */
    uint16      s_want_extra_isize;         /* New inodes should reserve # bytes */
    uint32      s_flags;                    /* Miscellaneous flags */
    uint16      s_raid_stride;              /* RAID stride */
    uint16      s_mmp_update_interval;      /* # seconds to wait in MMP checking */
    uint64      s_mmp_block;                /* Block for multi-mount protection */
    uint32      s_raid_stripe_width;        /* blocks on all data disks (N*stride)*/
    uint8       s_log_groups_per_flex;      /* FLEX_BG group size */
    uint8       s_checksum_type;            /* metadata checksum algorithm used */
    uint8       s_encryption_level;         /* versioning level for encryption */
    uint8       s_reserved_pad;             /* Padding to next 32bits */
    uint64      s_kbytes_written;           /* nr of lifetime kilobytes written */
    uint32      s_snapshot_inum;            /* Inode number of active snapshot */
    uint32      s_snapshot_id;              /* sequential ID of active snapshot */
    uint64      s_snapshot_r_blocks_count;  /* reserved blocks for active snapshot's future use */
    uint32      s_snapshot_list;            /* inode number of the head of the on-disk snapshot list */
    uint32      s_error_count;              /* number of fs errors */
    uint32      s_first_error_time;         /* first time an error happened */
    uint32      s_first_error_ino;          /* inode involved in first error */
    uint64      s_first_error_block;        /* block involved of first error */
    uint8       s_first_error_func[32];     /* function where the error happened */
    uint32      s_first_error_line;         /* line number where error happened */
    uint32      s_last_error_time;          /* most recent time of an error */
    uint32      s_last_error_ino;           /* inode involved in last error */
    uint32      s_last_error_line;          /* line number where error happened */
    uint64      s_last_error_block;         /* block involved of last error */
    uint8       s_last_error_func[32];      /* function where the error happened */
    uint8       s_mount_opts[64];
    uint32      s_usr_quota_inum;           /* inode for tracking user quota */
    uint32      s_grp_quota_inum;           /* inode for tracking group quota */
    uint32      s_overhead_clusters;        /* overhead blocks/clusters in fs */
    uint32      s_backup_bgs[2];            /* groups with sparse_super2 SBs */
    uint8       s_encrypt_algos[4];         /* Encryption algorithms in use  */
    uint8       s_encrypt_pw_salt[16];      /* Salt used for string2key algorithm */
    uint32      s_lpf_ino;                  /* Location of the lost+found inode */
    uint32      s_prj_quota_inum;           /* inode for tracking project quota */
    uint32      s_checksum_seed;            /* crc32c(uuid) if csum_seed set */
    uint32      s_reserved[98];             /* Padding to the end of the block */
    uint32      s_checksum;                 /* crc32c(superblock) */
};

struct ext2_group_desc
{
    uint32      bg_block_bitmap_lo;         /* Blocks bitmap block */
    uint32      bg_inode_bitmap_lo;         /* Inodes bitmap block */
    uint32      bg_inode_table_lo;          /* Inodes table block */
    uint16      bg_free_blocks_count_lo;    /* Free blocks count */
    uint16      bg_free_inodes_count_lo;    /* Free inodes count */
    uint16      bg_used_dirs_count_lo;      /* Directories count */
    uint16      bg_pad;
    uint32      bg_reserved[3];
};

struct ext4_group_desc
{
    uint32      bg_block_bitmap_lo;         /* Blocks bitmap block */
    uint32      bg_inode_bitmap_lo;         /* Inodes bitmap block */
    uint32      bg_inode_table_lo;          /* Inodes table block */
    uint16      bg_free_blocks_count_lo;    /* Free blocks count */
    uint16      bg_free_inodes_count_lo;    /* Free inodes count */
    uint16      bg_used_dirs_count_lo;      /* Directories count */
    uint16      bg_flags;                   /* EXT4_BG_flags (INODE_UNINIT, etc) */
    uint32      bg_exclude_bitmap_lo;       /* Exclude bitmap for snapshots */
    uint16      bg_block_bitmap_csum_lo;    /* crc32c(s_uuid+grp_num+bbitmap) LE */
    uint16      bg_inode_bitmap_csum_lo;    /* crc32c(s_uuid+grp_num+ibitmap) LE */
    uint16      bg_itable_unused_lo;        /* Unused inodes count */
    uint16      bg_checksum;                /* crc16(sb_uuid+group+desc) */
    uint32      bg_block_bitmap_hi;         /* Blocks bitmap block MSB */
    uint32      bg_inode_bitmap_hi;         /* Inodes bitmap block MSB */
    uint32      bg_inode_table_hi;          /* Inodes table block MSB */
    uint16      bg_free_blocks_count_hi;    /* Free blocks count MSB */
    uint16      bg_free_inodes_count_hi;    /* Free inodes count MSB */
    uint16      bg_used_dirs_count_hi;      /* Directories count MSB */
    uint16      bg_itable_unused_hi;        /* Unused inodes count MSB */
    uint32      bg_exclude_bitmap_hi;       /* Exclude bitmap block MSB */
    uint16      bg_block_bitmap_csum_hi;    /* crc32c(s_uuid+grp_num+bbitmap) BE */
    uint16      bg_inode_bitmap_csum_hi;    /* crc32c(s_uuid+grp_num+ibitmap) BE */
    uint32      bg_reserved;
};

struct ext4_inode {
    uint16      i_mode;                     /* File mode */
    uint16      i_uid;                      /* Low 16 bits of Owner Uid */
    uint32      i_size_lo;                  /* Size in bytes */
    uint32      i_atime;                    /* Access time */
    uint32      i_ctime;                    /* Inode Change time */
    uint32      i_mtime;                    /* Modification time */
    uint32      i_dtime;                    /* Deletion Time */
    uint16      i_gid;                      /* Low 16 bits of Group Id */
    uint16      i_links_count;              /* Links count */
    uint32      i_blocks_lo;                /* Blocks count */
    uint32      i_flags;                    /* File flags */
    uint32      i_reserved_1;
    char        i_block[60];                /* Pointers to blocks */
    uint32      i_generation;               /* File version (for NFS) */
    uint32      i_file_acl_lo;              /* File ACL */
    uint32      i_size_high;
    uint32      i_obso_faddr;               /* Obsoleted fragment address */
    uint16      i_blocks_high;              /* were l_i_reserved1 */
    uint16      i_file_acl_high;
    uint16      i_uid_high;                 /* these 2 fields */
    uint16      i_gid_high;                 /* were reserved2[0] */
    uint16      i_checksum_lo;              /* crc32c(uuid+inum+inode) LE */
    uint16      i_reserved;
    uint16      i_extra_isize;
    uint16      i_checksum_hi;              /* crc32c(uuid+inum+inode) BE */
    uint32      i_ctime_extra;              /* extra Change time       (nsec << 2 | epoch) */
    uint32      i_mtime_extra;              /* extra Modification time (nsec << 2 | epoch) */
    uint32      i_atime_extra;              /* extra Access time       (nsec << 2 | epoch) */
    uint32      i_crtime;                   /* File Creation time */
    uint32      i_crtime_extra;             /* extra FileCreationtime  (nsec << 2 | epoch) */
    uint32      i_version_hi;               /* high 32 bits for 64-bit version */
    uint32      i_projid;                   /* Project ID */
    char        i_extra[256 - 128 - i_extra_isize];
};

struct ext2_dir_entry {
    uint32      inode;                      /* Inode number */
    uint16      rec_len;                    /* Directory entry length */
    uint16      name_len;                   /* Name length */
    char        name[0];                    /* File name */
};

struct ext2_dir_entry_2 {
    uint32      inode;                      /* Inode number */
    uint16      rec_len;                    /* Directory entry length */
    uint8       name_len;                   /* Name length */
    uint8       file_type;
    char        name[0];                    /* File name */
};

struct ext4_dir_entry_tail {
    uint32      det_reserved_zero1;         /* Pretend to be unused */
    uint16      det_rec_len;                /* 12 */
    uint8       det_reserved_zero2;         /* Zero name length */
    uint8       det_reserved_ft;            /* 0xDE, fake file type */
    uint32      det_checksum;               /* crc32c(uuid+inum+dirblock) */
};

struct dx_root {
    ext2_dir_entry_2    dot;
    char                _pad0[3];
    ext2_dir_entry_2    dotdot;
    char                _pad1[2];
    uint32              reserved_zero;
    uint8               hash_version;
    uint8               info_length;
    uint8               indirect_levels;
    uint8               unused_flags;
    uint16              limit;
    uint16              count;
    uint32              block;
};

struct dx_node {
    ext2_dir_entry_2    fake_direntry;
    uint16              limit;
    uint16              count;
    int32               block;
};

struct dx_entry
{
    uint32 hash;
    uint32 block;
};

struct ext4_extent_header {
    uint16      eh_magic;                   /* probably will support different formats */
    uint16      eh_entries;                 /* number of valid entries */
    uint16      eh_max;                     /* capacity of store in entries */
    uint16      eh_depth;                   /* has tree real underlying blocks? */
    uint32      eh_generation;              /* generation of the tree */
};

struct ext4_extent_idx {
    uint32      ei_block;                   /* index covers logical blocks from 'block' */
    uint32      ei_leaf_lo;                 /* pointer to the physical block of the next level. */
                                            /* leaf or next index could be there */
    uint16      ei_leaf_hi;                 /* high 16 bits of physical block */
    uint16      ei_unused;
};

struct ext4_extent {
    uint32      ee_block;                   /* first logical block extent covers */
    uint16      ee_len;                     /* number of blocks covered by extent */
    uint16      ee_start_hi;                /* high 16 bits of physical block */
    uint32      ee_start_lo;                /* low 32 bits of physical block */
};

#define EXT4_XATTR_MAGIC                    0xEA020000

/* Name indexes */
#define EXT4_XATTR_INDEX_USER               1
#define EXT4_XATTR_INDEX_POSIX_ACL_ACCESS   2
#define EXT4_XATTR_INDEX_POSIX_ACL_DEFAULT  3
#define EXT4_XATTR_INDEX_TRUSTED            4
#define EXT4_XATTR_INDEX_LUSTRE             5
#define EXT4_XATTR_INDEX_SECURITY           6
#define EXT4_XATTR_INDEX_SYSTEM             7
#define EXT4_XATTR_INDEX_RICHACL            8
#define EXT4_XATTR_INDEX_ENCRYPTION         9
#define EXT4_XATTR_INDEX_HURD               10  /* Reserved for Hurd */

#define XATTR_HURD_PREFIX                   "gnu."
#define XATTR_SECURITY_PREFIX               "security."
#define XATTR_SYSTEM_PREFIX                 "system."
#define XATTR_TRUSTED_PREFIX                "trusted."
#define XATTR_USER_PREFIX                   "user."
#define XATTR_POSIX_ACL_ACCESS              "posix_acl_access"
#define XATTR_POSIX_ACL_DEFAULT             "posix_acl_default"

#define EXT4_XATTR_PAD_BITS                 2
#define EXT4_XATTR_PAD                      (1<<EXT4_XATTR_PAD_BITS)
#define EXT4_XATTR_ROUND                    (EXT4_XATTR_PAD-1)

struct ext4_xattr_header {
    uint32      h_magic;                    /* magic number for identification */
    uint32      h_refcount;                 /* reference count */
    uint32      h_blocks;                   /* number of disk blocks used */
    uint32      h_hash;                     /* hash value of all attributes */
    uint32      h_checksum;                 /* crc32c(uuid+id+xattrblock) */
                                            /* id = inum if refcount=1, blknum otherwise */
    uint32      h_reserved[3];              /* zero right now */
};

struct ext4_xattr_ibody_header {
    uint32      h_magic;                    /* magic number for identification */
};

struct ext4_xattr_entry {
    uint8       e_name_len;                 /* length of name */
    uint8       e_name_index;               /* attribute name index */
    uint16      e_value_offs;               /* offset in disk block of value */
    uint32      e_value_inum;               /* inode in which the value is stored */
    uint32      e_value_size;               /* size of attribute value */
    uint32      e_hash;                     /* hash value of name and value */
    char        e_name[e_name_len];         /* attribute name */
};
"""

c_ext = cstruct().load(ext_def)

EXT2 = 2
EXT3 = 3
EXT4 = 4

# Filetypes used in v2 directory listings
FILETYPES = {
    0x1: stat.S_IFREG,
    0x2: stat.S_IFDIR,
    0x3: stat.S_IFCHR,
    0x4: stat.S_IFBLK,
    0x5: stat.S_IFIFO,
    0x6: stat.S_IFSOCK,
    0x7: stat.S_IFLNK,
}


XATTR_PREFIX_MAP = {
    c_ext.EXT4_XATTR_INDEX_USER: c_ext.XATTR_USER_PREFIX,
    c_ext.EXT4_XATTR_INDEX_POSIX_ACL_ACCESS: c_ext.XATTR_SYSTEM_PREFIX,
    c_ext.EXT4_XATTR_INDEX_POSIX_ACL_DEFAULT: c_ext.XATTR_SYSTEM_PREFIX,
    c_ext.EXT4_XATTR_INDEX_TRUSTED: c_ext.XATTR_TRUSTED_PREFIX,
    c_ext.EXT4_XATTR_INDEX_SECURITY: c_ext.XATTR_SECURITY_PREFIX,
    c_ext.EXT4_XATTR_INDEX_SYSTEM: c_ext.XATTR_SYSTEM_PREFIX,
}

XATTR_NAME_MAP = {
    c_ext.EXT4_XATTR_INDEX_POSIX_ACL_ACCESS: c_ext.XATTR_POSIX_ACL_ACCESS,
    c_ext.EXT4_XATTR_INDEX_POSIX_ACL_DEFAULT: c_ext.XATTR_POSIX_ACL_DEFAULT,
}



from __future__ import annotations

from dissect.extfs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.extfs.extfs import ExtFS, INode
from dissect.extfs.journal import JDB2

__all__ = [
    "JDB2",
    "Error",
    "ExtFS",
    "FileNotFoundError",
    "INode",
    "NotADirectoryError",
    "NotASymlinkError",
]



from __future__ import annotations

import io
import logging
import os
import stat
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.util import ts
from dissect.util.stream import RangeStream, RunlistStream

from dissect.extfs.c_ext import (
    EXT2,
    EXT3,
    EXT4,
    FILETYPES,
    XATTR_NAME_MAP,
    XATTR_PREFIX_MAP,
    c_ext,
)
from dissect.extfs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.extfs.journal import JDB2

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_EXTFS", "CRITICAL"))


class ExtFS:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        fh.seek(c_ext.EXT2_SBOFF)
        sb = c_ext.ext4_super_block(fh)
        self.sb = sb

        if sb.s_magic != c_ext.EXT2_FS_MAGIC:
            raise Error("Not a valid ExtFS filesystem (magic mismatch)")

        if sb.s_inodes_count < 10:
            raise Error("Not a valid ExtFS filesystem (inum count < 10)")

        if sb.s_blocks_per_group == 0 or sb.s_inodes_per_group == 0:
            raise Error("Not a valid ExtFS filesystem (blocks or inodes per group is 0)")

        if sb.s_log_block_size != sb.s_log_cluster_size:
            raise NotImplementedError("Different size cluster than blocks is currently not supported")

        self.block_size = c_ext.EXT2_MIN_BLOCK_SIZE << sb.s_log_block_size
        if self.block_size == 0 or self.block_size % 512:
            raise Error("Not a valid ExtFS filesystem (invalid block size)")

        if sb.s_feature_incompat & c_ext.EXT4_FEATURE_INCOMPAT_EXTENTS:
            self.type = EXT4
        elif sb.s_feature_compat & c_ext.EXT3_FEATURE_COMPAT_HAS_JOURNAL:
            self.type = EXT3
        else:
            self.type = EXT2

        if sb.s_feature_incompat & c_ext.EXT2_FEATURE_INCOMPAT_FILETYPE:
            self._dirtype = c_ext.ext2_dir_entry_2
        else:
            self._dirtype = c_ext.ext2_dir_entry

        self.block_count = (sb.s_blocks_count_hi << 32) | sb.s_blocks_count_lo
        self.last_block = self.block_count - 1

        if (
            self.type == EXT4
            and self.sb.s_feature_incompat & c_ext.EXT4_FEATURE_INCOMPAT_64BIT
            and self.sb.s_desc_size >= 64
        ):
            self._group_desc_struct = c_ext.ext4_group_desc
        else:
            self._group_desc_struct = c_ext.ext2_group_desc
        self._group_desc_size = sb.s_desc_size if sb.s_desc_size else len(self._group_desc_struct)

        goff = c_ext.EXT2_SBOFF + self._group_desc_size
        self.groups_offset = goff if goff % self.block_size == 0 else goff + self.block_size - goff % self.block_size
        self.groups_count = ((self.last_block - sb.s_first_data_block) // sb.s_blocks_per_group) + 1

        self.uuid = UUID(bytes=sb.s_uuid)
        self.volume_name = sb.s_volume_name.split(b"\x00")[0].decode(errors="surrogateescape")
        self.last_mount = sb.s_last_mounted.split(b"\x00")[0].decode(errors="surrogateescape")

        self.root = self.get_inode(c_ext.EXT2_ROOT_INO, "/")

        self.get_inode = lru_cache(1024)(self.get_inode)
        self._read_group_desc = lru_cache(356)(self._read_group_desc)

    @cached_property
    def journal(self) -> JDB2:
        if not self.sb.s_feature_compat & c_ext.EXT3_FEATURE_COMPAT_HAS_JOURNAL:
            raise Error("Journal not supported")

        inum = self.sb.s_journal_inum
        if inum == 0:
            raise Error(f"Journal inum is 0, could be on external device (s_journal_uuid = {self.sb.s_journal_uuid})")

        inode = self.get_inode(inum)
        return JDB2(inode.open())

    def get(self, path_or_inum: str | int, node: INode | None = None) -> INode:
        if isinstance(path_or_inum, int):
            return self.get_inode(path_or_inum)

        node = node if node else self.root
        parts = path_or_inum.split("/")
        for part in parts:
            if not part:
                continue

            for entry in node.iterdir():
                if entry.filename == part:
                    node = entry
                    break
            else:
                raise FileNotFoundError(f"File not found: {path_or_inum}")

        return node

    def get_inode(
        self,
        inum: int,
        filename: str | None = None,
        filetype: int | None = None,
    ) -> INode:
        if inum < c_ext.EXT2_BAD_INO or inum > self.sb.s_inodes_count:
            raise Error(f"inum out of range {c_ext.EXT2_BAD_INO}-{self.sb.s_inodes_count}: {inum}")

        return INode(self, inum, filename, filetype)

    def _read_group_desc(self, group_num: int) -> c_ext.ext2_group_desc | c_ext.ext4_group_desc:
        if group_num >= self.groups_count:
            raise Error("Group number exceeds amount of groups")

        offset = self.groups_offset + group_num * self._group_desc_size
        self.fh.seek(offset)
        group_desc = self._group_desc_struct(self.fh)

        if self._group_desc_struct == c_ext.ext4_group_desc:
            block_bitmap = (group_desc.bg_block_bitmap_hi << 32) | group_desc.bg_block_bitmap_lo
            inode_bitmap = (group_desc.bg_inode_bitmap_hi << 32) | group_desc.bg_inode_bitmap_lo
            table_block = (group_desc.bg_inode_table_hi << 32) | group_desc.bg_inode_table_lo
        else:
            block_bitmap = group_desc.bg_block_bitmap_lo
            inode_bitmap = group_desc.bg_inode_bitmap_lo
            table_block = group_desc.bg_inode_table_lo

        if block_bitmap > self.last_block or inode_bitmap > self.last_block or table_block > self.last_block:
            raise Error("Group descriptor block locations exceed last block")

        return group_desc


class INode:
    def __init__(
        self,
        extfs: ExtFS,
        inum: int,
        filename: str | None = None,
        filetype: int | None = None,
    ):
        self.extfs = extfs
        self.inum = inum
        self.filename = filename
        self._filetype = filetype
        self._runlist = None

    def __repr__(self) -> str:
        return f"<inode {self.inum}>"

    @cached_property
    def inode(self) -> c_ext.ext4_inode:
        block_group_num, index = divmod(self.inum - 1, self.extfs.sb.s_inodes_per_group)
        block_group = self.extfs._read_group_desc(block_group_num)

        if self.extfs._group_desc_struct == c_ext.ext4_group_desc:
            table_block = (block_group.bg_inode_table_hi << 32) | block_group.bg_inode_table_lo
        else:
            table_block = block_group.bg_inode_table_lo

        offset = table_block * self.extfs.block_size + index * self.extfs.sb.s_inode_size
        self.extfs.fh.seek(offset)
        return c_ext.ext4_inode(self.extfs.fh)

    @cached_property
    def size(self) -> int:
        return (self.inode.i_size_high << 32) + self.inode.i_size_lo

    @property
    def filetype(self) -> int:
        if not self._filetype:
            self._filetype = stat.S_IFMT(self.inode.i_mode)
        return self._filetype

    @cached_property
    def link(self) -> str:
        if self.filetype != stat.S_IFLNK:
            raise NotASymlinkError(f"{self!r} is not a symlink")

        return self.open().read().decode(errors="surrogateescape")

    @cached_property
    def xattr(self) -> list[XAttr]:
        xattr = []

        if self.inode.i_extra.strip(b"\x00"):
            buf = io.BytesIO(self.inode.i_extra)
            hdr = c_ext.ext4_xattr_ibody_header(buf)
            if hdr.h_magic != c_ext.EXT4_XATTR_MAGIC:
                raise Error("Invalid xattr magic value")

            xattr.extend(_iter_xattr(self, buf, len(self.inode.i_extra), 4))

        if self.inode.i_file_acl_lo:
            block = (self.inode.i_file_acl_high << 32) | self.inode.i_file_acl_lo
            block_offset = block * self.extfs.block_size

            buf = RangeStream(self.extfs.fh, block_offset, self.extfs.block_size)
            hdr = c_ext.ext4_xattr_header(buf)
            if hdr.h_magic != c_ext.EXT4_XATTR_MAGIC:
                raise Error("Invalid xattr magic value")

            xattr.extend(_iter_xattr(self, buf, buf.size))

        return xattr

    @property
    def atime(self) -> datetime:
        return ts.from_unix_ns(self.atime_ns)

    @property
    def atime_ns(self) -> int:
        time = self.inode.i_atime
        time_extra = self.inode.i_atime_extra if self.extfs.sb.s_inode_size > 128 else 0

        return _parse_ns_ts(time, time_extra)

    @property
    def mtime(self) -> datetime:
        return ts.from_unix_ns(self.mtime_ns)

    @property
    def mtime_ns(self) -> int:
        time = self.inode.i_mtime
        time_extra = self.inode.i_mtime_extra if self.extfs.sb.s_inode_size > 128 else 0

        return _parse_ns_ts(time, time_extra)

    @property
    def ctime(self) -> datetime:
        return ts.from_unix_ns(self.ctime_ns)

    @property
    def ctime_ns(self) -> int:
        time = self.inode.i_ctime
        time_extra = self.inode.i_ctime_extra if self.extfs.sb.s_inode_size > 128 else 0

        return _parse_ns_ts(time, time_extra)

    @property
    def dtime(self) -> datetime:
        return ts.from_unix(self.inode.i_dtime)

    @property
    def crtime(self) -> datetime | None:
        time_ns = self.crtime_ns
        if time_ns is None:
            return None
        return ts.from_unix_ns(time_ns)

    @property
    def crtime_ns(self) -> int | None:
        if self.extfs.sb.s_inode_size <= 128:
            return None

        time = self.inode.i_crtime
        time_extra = self.inode.i_crtime_extra

        return _parse_ns_ts(time, time_extra)

    def listdir(self) -> dict[str, INode]:
        return {node.filename: node for node in self.iterdir()}

    dirlist = listdir

    def iterdir(self) -> Iterator[INode]:
        if self.filetype != stat.S_IFDIR:
            raise NotADirectoryError(f"{self!r} is not a directory")

        buf = self.open()
        offset = 0

        while offset < self.size - 12:
            direntry = self.extfs._dirtype(buf)

            if direntry.rec_len == 0:
                log.critical("Zero-length directory entry in %s (offset 0x%x)", self, offset)
                return

            # Sanity check if the direntry is valid
            if 0 < direntry.inode < self.extfs.sb.s_inodes_count:
                fname = buf.read(direntry.name_len).decode(errors="surrogateescape")
                ftype = direntry.file_type if self.extfs._dirtype == c_ext.ext2_dir_entry_2 else None

                if ftype:
                    ftype = FILETYPES[ftype]

                yield self.extfs.get_inode(direntry.inode, fname, ftype)

            offset += direntry.rec_len
            buf.seek(offset)

    def dataruns(self) -> list[tuple[int | None, int]]:
        if not self._runlist:
            expected_runs = (self.size + self.extfs.block_size - 1) // self.extfs.block_size

            if self.inode.i_flags & c_ext.EXT4_EXTENTS_FL:
                buf = io.BytesIO(self.inode.i_block)

                runs = []
                run_offset = 0

                for extent in _parse_extents(self, buf):
                    # Account for uninitialized extents
                    if extent.ee_len > 0x8000:
                        uninitialized_gap = extent.ee_len - 0x8000
                        runs.append((None, uninitialized_gap))
                        run_offset += uninitialized_gap
                        continue

                    # Account for sparse gaps
                    if extent.ee_block != run_offset:
                        sparse_gap = extent.ee_block - run_offset
                        runs.append((None, sparse_gap))
                        run_offset += sparse_gap

                    runs.append(((extent.ee_start_hi << 32) | extent.ee_start_lo, extent.ee_len))
                    run_offset += extent.ee_len

                if run_offset < expected_runs:
                    runs.append((None, expected_runs - run_offset))

                self._runlist = runs
            else:
                i_blocks = c_ext.uint32[15](self.inode.i_block)
                num_blocks = (self.size + self.extfs.block_size - 1) // self.extfs.block_size
                num_direct_blocks = min(num_blocks, c_ext.EXT2_NDIR_BLOCKS)

                blocks = i_blocks[:num_direct_blocks]
                num_blocks -= num_direct_blocks

                if num_blocks > 0:
                    for level in range(c_ext.EXT2_NIND_BLOCKS):
                        indirect_offset = i_blocks[num_direct_blocks + level]
                        parsed_blocks = _parse_indirect(self, indirect_offset, num_blocks, level + 1)
                        num_blocks -= len(parsed_blocks)
                        blocks.extend(parsed_blocks)

                        if num_blocks == 0:
                            break

                runs = []
                if blocks:
                    run_offset = None
                    run_size = 1

                    for block in blocks:
                        if run_offset is None:
                            run_offset = block
                            continue

                        if block == run_offset + run_size:
                            run_size += 1
                        else:
                            if run_offset == 0:
                                runs.append((None, run_size))
                            else:
                                runs.append((run_offset, run_size))
                            run_offset = block
                            run_size = 1

                    runs.append((run_offset, run_size))

                self._runlist = runs

        return self._runlist

    def open(self) -> BinaryIO:
        if self.inode.i_flags & c_ext.EXT4_INLINE_DATA_FL or (self.filetype == stat.S_IFLNK and self.size < 60):
            buf = io.BytesIO(memoryview(self.inode.i_block)[: self.size])
            # Need to add a size attribute to maintain compatibility with dissect streams
            buf.size = self.size
            return buf
        return RunlistStream(self.extfs.fh, self.dataruns(), self.size, self.extfs.block_size)


class XAttr:
    def __init__(self, extfs: ExtFS, inode: INode, entry: c_ext.ext4_xattr_entry, value: bytes):
        self.extfs = extfs
        self.inode = inode
        self.entry = entry

        self.prefix = XATTR_PREFIX_MAP.get(entry.e_name_index, "unknown_prefix")
        self._name = XATTR_NAME_MAP.get(entry.e_name_index, entry.e_name.decode(errors="surrogateescape"))
        self.name = self.prefix + self._name
        self.value = value

    def __repr__(self) -> str:
        return f"<xattr name={self.name} value={self.value} inode={self.inode}>"


def _parse_indirect(inode: INode, offset: int, num_blocks: int, level: int) -> list[int]:
    offsets_per_block = inode.extfs.block_size // 4

    if level == 1:
        read_blocks = min(num_blocks, offsets_per_block)
        if offset == 0:
            return [0] * read_blocks
        inode.extfs.fh.seek(offset * inode.extfs.block_size)
        return c_ext.uint32[read_blocks](inode.extfs.fh)

    blocks = []

    max_level_blocks = offsets_per_block**level
    blocks_per_nest = max_level_blocks // offsets_per_block
    read_blocks = (num_blocks + blocks_per_nest - 1) // blocks_per_nest
    read_blocks = min(read_blocks, offsets_per_block)

    inode.extfs.fh.seek(offset * inode.extfs.block_size)
    for addr in c_ext.uint32[read_blocks](inode.extfs.fh):
        parsed_blocks = _parse_indirect(inode, addr, num_blocks, level - 1)
        num_blocks -= len(parsed_blocks)
        blocks.extend(parsed_blocks)

    return blocks


def _parse_extents(inode: INode, buf: bytes) -> Iterator[c_ext.ext4_extent]:
    extent_header = c_ext.ext4_extent_header(buf)

    if extent_header.eh_magic != 0xF30A:
        raise Error("Invalid extent_header magic")

    if extent_header.eh_depth == 0:
        for _ in range(extent_header.eh_entries):
            extent = c_ext.ext4_extent(buf)
            yield extent
    else:
        for _ in range(extent_header.eh_entries):
            idx = c_ext.ext4_extent_idx(buf)
            child = (idx.ei_leaf_hi << 32) | idx.ei_leaf_lo

            fh = inode.extfs.fh
            fh.seek(child * inode.extfs.block_size)
            blockbuf = io.BytesIO(fh.read(inode.extfs.block_size))
            yield from _parse_extents(inode, blockbuf)


def _iter_xattr(inode: INode, buf: BinaryIO, end: int, value_offset: int = 0) -> Iterator[XAttr]:
    offset = buf.tell()
    while True:
        try:
            if offset > end:
                break

            buf.seek(offset)
            entry = c_ext.ext4_xattr_entry(buf)

            if (entry.e_name_len, entry.e_name_index, entry.e_value_offs) == (0, 0, 0):
                break

            if entry.e_value_inum:
                value = inode.extfs.get_inode(entry.e_value_inum).open().read(entry.e_value_size)
            else:
                buf.seek(value_offset + entry.e_value_offs)
                value = buf.read(entry.e_value_size)

            yield XAttr(inode.extfs, inode, entry, value)

            offset += (len(entry) + c_ext.EXT4_XATTR_ROUND) & (~c_ext.EXT4_XATTR_ROUND & 0xFFFFFFFF)
        except EOFError:
            break


def _parse_ns_ts(time: int, time_extra: int) -> int:
    # The low 2 bits of time_extra are used to extend the time field
    # The remaining 30 bits are nanoseconds
    time |= (time_extra & 0b11) << 32
    ns = time_extra >> 2

    return (time * 1000000000) + ns



from __future__ import annotations

from dissect.cstruct import cstruct

jdb2_def = """
#define JBD2_MAGIC_NUMBER                   0xC03B3998

#define JBD2_DESCRIPTOR_BLOCK               1
#define JBD2_COMMIT_BLOCK                   2
#define JBD2_SUPERBLOCK_V1                  3
#define JBD2_SUPERBLOCK_V2                  4
#define JBD2_REVOKE_BLOCK                   5

#define JBD2_CRC32_CHKSUM                   1
#define JBD2_MD5_CHKSUM                     2
#define JBD2_SHA1_CHKSUM                    3
#define JBD2_CRC32C_CHKSUM                  4
#define JBD2_CRC32_CHKSUM_SIZE              4

#define JBD2_FLAG_ESCAPE                    1   // on-disk block is escaped
#define JBD2_FLAG_SAME_UUID                 2   // block has same uuid as previous
#define JBD2_FLAG_DELETED                   4   // block deleted by this transaction
#define JBD2_FLAG_LAST_TAG                  8   // last tag in this descriptor block

#define JBD2_FEATURE_COMPAT_CHECKSUM        0x00000001

#define JBD2_FEATURE_INCOMPAT_REVOKE        0x00000001
#define JBD2_FEATURE_INCOMPAT_64BIT         0x00000002
#define JBD2_FEATURE_INCOMPAT_ASYNC_COMMIT  0x00000004
#define JBD2_FEATURE_INCOMPAT_CSUM_V2       0x00000008
#define JBD2_FEATURE_INCOMPAT_CSUM_V3       0x00000010

struct journal_header {
    uint32  h_magic;
    uint32  h_blocktype;
    uint32  h_sequence;
};

struct journal_superblock {
    journal_header s_header;
    /* Static information describing the journal */
    uint32  s_blocksize;            /* journal device blocksize */
    uint32  s_maxlen;               /* total blocks in journal file */
    uint32  s_first;                /* first block of log information */
    /* Dynamic information describing the current state of the log */
    uint32  s_sequence;             /* first commit ID expected in log */
    uint32  s_start;                /* blocknr of start of log */
    uint32  s_errno;
    /* Remaining fields are only valid in a version-2 superblock */
    uint32  s_feature_compat;       /* compatible feature set */
    uint32  s_feature_incompat;     /* incompatible feature set */
    uint32  s_feature_ro_compat;    /* readonly-compatible feature set */
    char    s_uuid[16];             /* 128-bit uuid for journal */
    uint32  s_nr_users;             /* Nr of filesystems sharing log */
    uint32  s_dynsuper;             /* Blocknr of dynamic superblock copy*/
    uint32  s_max_transaction;      /* Limit of journal blocks per trans.*/
    uint32  s_max_trans_data;       /* Limit of data blocks per trans. */
    uint8   s_checksum_type;        /* checksum type */
    char    s_padding2[3];
    char    s_padding[168];
    uint32  s_checksum;             /* crc32c(superblock) */
    uint8   s_users[16*48];         /* ids of all fs'es sharing the log */
};

struct commit_header {
    uint32  h_magic;
    uint32  h_blocktype;
    uint32  h_sequence;
    uint8   h_chksum_type;
    uint8   h_chksum_size;
    char    h_padding[2];
    char    h_chksum[32];
    uint64  h_commit_sec;
    uint32  h_commit_nsec;
};

struct journal_block_tag {
    uint32  t_blocknr;          /* The on-disk block number */
    uint16  t_checksum;         /* truncated crc32c(uuid+seq+block) */
    uint16  t_flags;            /* See below */
    uint32  t_blocknr_high;     /* most-significant high 32bits. */
};

struct journal_block_tag3 {
    uint32  t_blocknr;          /* The on-disk block number */
    uint32  t_flags;            /* See below */
    uint32  t_blocknr_high;     /* most-significant high 32bits. */
    uint32  t_checksum;         /* crc32c(uuid+seq+block) */
};
"""

c_jdb2 = cstruct(endian=">").load(jdb2_def)



from __future__ import annotations

import datetime
import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import RangeStream

from dissect.extfs.c_jdb2 import c_jdb2
from dissect.extfs.exceptions import Error

if TYPE_CHECKING:
    from collections.abc import Iterator


class JDB2:
    def __init__(self, fh: BinaryIO):
        self.fh = fh

        sb = c_jdb2.journal_superblock(self.fh)
        self.sb = sb

        if sb.s_header.h_magic != c_jdb2.JBD2_MAGIC_NUMBER:
            raise Error("Not a valid JDB2 journal (magic mismatch)")

        self.block_size = sb.s_blocksize
        self._blocktag = c_jdb2.journal_block_tag
        if sb.s_feature_incompat & c_jdb2.JBD2_FEATURE_INCOMPAT_CSUM_V3:
            self._blocktag = c_jdb2.journal_block_tag3

    def read_block(self, block: int, count: int = 1) -> bytes:
        offset = block * self.block_size
        self.fh.seek(offset)
        return self.fh.read(self.block_size * count)

    def commits(self) -> Iterator[CommitBlock]:
        cur_seq = None

        for block in self.commits_all():
            if not cur_seq:
                cur_seq = block.sequence

            if block.sequence == cur_seq:
                cur_seq += 1
                yield block

    def commits_all(self) -> Iterator[CommitBlock]:
        desc_buf = {}

        for block in self.walk():
            if isinstance(block, DescriptorBlock):
                if block.sequence not in desc_buf:
                    desc_buf[block.sequence] = []

                desc_buf[block.sequence].append(block)
            elif isinstance(block, CommitBlock):
                block.descriptors = desc_buf.get(block.sequence, [])

                if block.sequence in desc_buf:
                    del desc_buf[block.sequence]

                yield block

    def walk(self) -> Iterator[CommitBlock]:
        block_num = self.sb.s_first

        while block_num < self.sb.s_maxlen - 1:
            offset = block_num * self.block_size
            self.fh.seek(offset)

            header = c_jdb2.journal_header(self.fh)
            if header.h_magic != c_jdb2.JBD2_MAGIC_NUMBER:
                block_num += 1
                continue

            if header.h_blocktype == c_jdb2.JBD2_DESCRIPTOR_BLOCK:
                yield DescriptorBlock(self, header, block_num)
            elif header.h_blocktype == c_jdb2.JBD2_COMMIT_BLOCK:
                self.fh.seek(offset)
                yield CommitBlock(self, c_jdb2.commit_header(self.fh), block_num)
            elif header.h_blocktype == c_jdb2.JBD2_REVOKE_BLOCK:
                pass

            block_num += 1


class DescriptorBlock:
    def __init__(self, jdb2: JDB2, header: c_jdb2.journal_header, block: int):
        self.jdb2 = jdb2
        self.header = header
        self.journal_block = block

        self.sequence = self.header.h_sequence

    def __repr__(self) -> str:
        return f"<descriptor_block sequence={self.sequence} journal_block={self.journal_block}>"

    def tags(self) -> Iterator[DescriptorBlockTag]:
        self.jdb2.fh.seek((self.journal_block * self.jdb2.block_size) + c_jdb2.journal_header.size)

        block_count = 1
        while True:
            tag = self.jdb2._blocktag(self.jdb2.fh)
            yield DescriptorBlockTag(self, tag, self.journal_block + block_count)

            if tag.t_flags & c_jdb2.JBD2_FLAG_LAST_TAG:
                break

            if not tag.t_flags & c_jdb2.JBD2_FLAG_SAME_UUID:
                self.jdb2.fh.seek(16, io.SEEK_CUR)
            block_count += 1


class DescriptorBlockTag:
    def __init__(
        self, descriptor: DescriptorBlock, tag: c_jdb2.journal_block_tag | c_jdb2.journal_block_tag3, journal_block: int
    ):
        self.descriptor = descriptor
        self.tag = tag
        self.journal_block = journal_block

        self.block = (self.tag.t_blocknr_high << 32) | self.tag.t_blocknr

    def __repr__(self) -> str:
        return f"<block_tag block={self.block} journal_block={self.journal_block} flags=0x{self.tag.t_flags:x}>"

    def open(self) -> BinaryIO:
        block_size = self.descriptor.jdb2.block_size
        return RangeStream(self.descriptor.jdb2.fh, self.journal_block * block_size, block_size)


class CommitBlock:
    def __init__(
        self,
        jdb2: JDB2,
        header: c_jdb2.commit_header,
        journal_block: int,
        descriptors: list[DescriptorBlock] | None = None,
    ):
        self.jdb2 = jdb2
        self.header = header
        self.journal_block = journal_block
        self.descriptors = descriptors if descriptors else []

        self.sequence = self.header.h_sequence
        self.ts = datetime.datetime.fromtimestamp(self.header.h_commit_sec, tz=datetime.timezone.utc)
        self.ts += datetime.timedelta(microseconds=self.header.h_commit_nsec // 1000)

    def __repr__(self) -> str:
        return (
            f"<commit sequence={self.sequence} journal_block={self.journal_block} "
            f"ts={self.ts} num_descriptors={len(self.descriptors)}>"
        )
